package jebl.evolution.align;

public class SystemOut extends Output {

    public final void print(String s) { System.out.print(s); }
    public final void println(String s) { System.out.println(s); }
    public final void println() { System.out.println(); }
}
